import 'package:flutter/material.dart';
import 'package:flutter/services.dart';

typedef ValueTransformer<T> = dynamic Function(T value);

class FormBuilder extends StatefulWidget {
  //final BuildContext context;
  final Function(Map<String, dynamic>) onChanged;
  final WillPopCallback onWillPop;
  final Widget child;
  final bool readOnly;
  final bool autovalidate;
  final Map<String, dynamic> initialValue;

  const FormBuilder({
    Key key,
    @required this.child,
    this.readOnly = false,
    this.onChanged,
    this.autovalidate = false,
    this.onWillPop,
    this.initialValue = const {},
  }) : super(key: key);

  static FormBuilderState of(BuildContext context) =>
      context.findAncestorStateOfType<FormBuilderState>();

  @override
  FormBuilderState createState() => FormBuilderState();
}

class FormBuilderState extends State<FormBuilder> {
  //TODO: Find way to assert no duplicates in field attributes
  final GlobalKey<FormState> _formKey = GlobalKey<FormState>();

  Map<String, GlobalKey<FormFieldState>> _fieldKeys;

  Map<String, dynamic> _value;

  Map<String, dynamic> get value => {...widget.initialValue ?? {}, ..._value};

  Map<String, dynamic> get initialValue => widget.initialValue;

  Map<String, GlobalKey<FormFieldState>> get fields => _fieldKeys;

  List<FocusNode> _fieldFocusNodes;
  Map<String, int> _attrIndex;

  int _currentFocusIndex;

  int _currentFocusCount = 0;

  bool get readOnly => widget.readOnly;

  @override
  void initState() {
    _fieldKeys = {};
    _value = {};
    _fieldFocusNodes = [];
    _attrIndex = {};
    super.initState();
  }

  @override
  void dispose() {
    _fieldKeys = null;
    super.dispose();
  }

  void setAttributeValue(String attribute, dynamic value) {
    setState(() {
      _value[attribute] = value;
    });
  }

  registerFieldKey(String attribute, GlobalKey key) {
    // assert(_fieldKeys.containsKey(attribute) == false, "Field with attribute '$attribute' already exists. Make sure that two or more fields don't have the same attribute name.");
    this._fieldKeys[attribute] = key;
  }

  registerFieldFocus(String attribute, FocusNode focus) {
    this._attrIndex[attribute] = _currentFocusCount;
    this._fieldFocusNodes.add(focus);
    _currentFocusCount++;
  }

  unregisterFieldKey(String attribute) {
    this._fieldKeys.remove(attribute);
  }

  void setCurrentFocus(String attr) {
    _currentFocusIndex = _attrIndex[attr];
  }

  FocusNode getCurrentFocusNode() {
    FocusNode node;
    try {
      node = _fieldFocusNodes[_currentFocusIndex];
    } catch (ex) {
      node = null;
    }
    return node;
  }

  void nextFocus() {
    if (_currentFocusIndex == _currentFocusCount) {
      getCurrentFocusNode().unfocus();
      return;
    }
    _currentFocusIndex++;
    getCurrentFocusNode().requestFocus();
  }

  /*changeAttributeValue(String attribute, dynamic newValue) {
    print(this.fieldKeys[attribute]);
    if (this.fieldKeys[attribute] != null){
      print("Current $attribute value: ${this.fieldKeys[attribute].currentState.value}");
      print("Trying to change $attribute to $newValue");
      this.fieldKeys[attribute].currentState.didChange(newValue);
      print("$attribute value after: ${this.fieldKeys[attribute].currentState.value}");
    }
  }*/

  void save() {
    _formKey.currentState.save();
  }

  bool validate() {
    return _formKey.currentState.validate();
  }

  bool saveAndValidate() {
    _formKey.currentState.save();
    return _formKey.currentState.validate();
  }

  void reset() {
    // _formKey.currentState.reset();
    _fieldKeys.forEach((mapKey, fieldKey) {
      // print("Reseting $mapKey");
      fieldKey.currentState.reset();
    });
  }

  @override
  Widget build(BuildContext context) {
    return Form(
      key: _formKey,
      child: widget.child,
      autovalidate: widget.autovalidate,
      onWillPop: widget.onWillPop,
      onChanged: () {
        if (widget.onChanged != null) {
          save();
          widget.onChanged(value);
        }
      },
    );
  }
}
